/* -*- Mode: C; tab-width: 8; indent-tabs-mode: nil; c-basic-offset: 4 -*-
 *
 * ***** BEGIN LICENSE BLOCK *****
 * Version: MPL 1.1/GPL 2.0/LGPL 2.1
 *
 * The contents of this file are subject to the Mozilla Public License Version
 * 1.1 (the "License"); you may not use this file except in compliance with
 * the License. You may obtain a copy of the License at
 * http://www.mozilla.org/MPL/
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * The Original Code is Mozilla Communicator client code, released
 * March 31, 1998.
 *
 * The Initial Developer of the Original Code is
 * Netscape Communications Corporation.
 * Portions created by the Initial Developer are Copyright (C) 1998
 * the Initial Developer. All Rights Reserved.
 *
 * Contributor(s):
 *
 * Alternatively, the contents of this file may be used under the terms of
 * either of the GNU General Public License Version 2 or later (the "GPL"),
 * or the GNU Lesser General Public License Version 2.1 or later (the "LGPL"),
 * in which case the provisions of the GPL or the LGPL are applicable instead
 * of those above. If you wish to allow use of your version of this file only
 * under the terms of either the GPL or the LGPL, and not to allow others to
 * use your version of this file under the terms of the MPL, indicate your
 * decision by deleting the provisions above and replace them with the notice
 * and other provisions required by the GPL or the LGPL. If you do not delete
 * the provisions above, a recipient may use your version of this file under
 * the terms of any one of the MPL, the GPL or the LGPL.
 *
 * ***** END LICENSE BLOCK ***** */

#include <stdarg.h>
#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <assert.h>
#define JSDMALLOC_C
#include "jsdmalloc.h"

/* Alignment is the guaranteed alignment of the memory pointer returned by
   js_dmalloc/js_dcalloc/js_drealloc. The system malloc must return pointers
   aligned to a multiple of this value to ensure correct behaviour. */
#define ALIGNMENT   (sizeof(double))
#define ALIGN(x)    (((x) + ALIGNMENT - 1) & ~(ALIGNMENT - 1))
#define GUARD_BYTES ALIGN(32)
#define PRE_SIZE    ALIGN(sizeof(Header) + GUARD_BYTES)
#define MARKER      0xA9B8C7D6
#define PRE_GUARD   0x53
#define ALLOC_FILL  0x55
#define POST_GUARD  0x57
#define FREE_FILL   0x59

/* Header pre-pended to a memory block */
typedef struct Header Header;
struct Header {
    unsigned    marker;
    const char *alloc_file;
    int         alloc_line;
    size_t      size;
    Header     *prev;
    Header     *next;
};

static Header *g_head;                     /* currently allocated blocks   */
static size_t  g_alloc_limit = (size_t)-1; /* when to start failing allocs */
static size_t  g_alloc_requests;           /* allocations requested        */
static size_t  g_alloc_bytes;              /* allocated bytes requested    */
static size_t  g_alloc_blocks;             /* number of allocated blocks   */
static size_t  g_corrupt_blocks;           /* corrupted blocks             */

/* Debug logging */
static void
dlog(const char *fmt, ...)
{
    va_list ap;

    va_start(ap, fmt);
    vfprintf(stdout, fmt, ap);
    va_end(ap);
    fflush(stdout);
}

/* Called whenever a failure is detected. */
static void
failure_detected(void)
{
}

/* Check whether a memory block is valid */
static int
check_block(const Header *hdr, const char *file, int line)
{
    unsigned char *ptr;
    size_t size;
    int result = 1;

    if (hdr->marker != MARKER) {
        dlog("Invalid block %p at %s:%d\n",
             (char*)hdr + PRE_SIZE, file, line);
        g_corrupt_blocks++;
        failure_detected();
        return 0;
    }

    ptr = (char*)hdr + PRE_SIZE - GUARD_BYTES;
    for (size = 0; size < GUARD_BYTES; size++) {
        if (*ptr++ != PRE_GUARD) {
            break;
        }
    }

    if (size != GUARD_BYTES) {
        dlog("Corrupted pre guard word for block %p %s:%d at %s:%d\n",
             (char*)hdr + PRE_SIZE, hdr->alloc_file, hdr->alloc_line,
             file, line);
        g_corrupt_blocks++;
        result = 0;
        failure_detected();
    }

    ptr = (char*)hdr + PRE_SIZE + hdr->size;
    for (size = 0; size < GUARD_BYTES; size++) {
        if (*ptr++ != POST_GUARD) {
            break;
        }
    }

    if (size != GUARD_BYTES) {
        dlog("Corrupted post guard word for block %p %s:%d at %s:%d\n",
             (char*)hdr + PRE_SIZE, hdr->alloc_file, hdr->alloc_line,
             file, line);
        g_corrupt_blocks++;
        result = 0;
        failure_detected();
    }

    return result;
}

void
js_dfree(void *ptr, const char *file, int line)
{
    Header *hdr;

    if (!ptr)
        return;

    hdr = (Header*)((char*)ptr - PRE_SIZE);
    if (!check_block(hdr, file, line))
        return;

    if (hdr->prev) {
        if (!check_block(hdr->prev, file, line))
            return;

        hdr->prev->next = hdr->next;
    } else {
        g_head = hdr->next;
    }

    if (hdr->next) {
        if (!check_block(hdr->next, file, line))
            return;

        hdr->next->prev = hdr->prev;
    }

    g_alloc_blocks--;
    g_alloc_bytes -= hdr->size;

    memset(hdr, FREE_FILL, PRE_SIZE + hdr->size + GUARD_BYTES);
    free(hdr);
}

void *
js_dmalloc(size_t size, const char *file, int line)
{
    Header *hdr;

    if (size == 0) {
        dlog("Attempt to allocate zero bytes from %s:%d\n",
             file, line);
        return NULL;
    }

    ++g_alloc_requests;
    if (g_alloc_requests > g_alloc_limit) {
        dlog("Limit %d reached. Failing alloc from %s:%d\n",
             g_alloc_limit, file, line);
        return NULL;
    }

    hdr = malloc(PRE_SIZE + size + GUARD_BYTES);
    if (!hdr) {
        dlog("Malloc of %d bytes failed at %s:%d\n",
             size, file, line);
        return NULL;
    }

    if (g_head)
        g_head->prev = hdr;

    hdr->marker     = MARKER;
    hdr->alloc_file = file;
    hdr->alloc_line = line;
    hdr->size       = size;
    hdr->next       = g_head;
    hdr->prev       = NULL;

    g_head          = hdr;
    g_alloc_bytes  += size;
    g_alloc_blocks++;

    memset((char*)hdr + PRE_SIZE - GUARD_BYTES, PRE_GUARD, GUARD_BYTES);
    memset((char*)hdr + PRE_SIZE + size, POST_GUARD, GUARD_BYTES);
    memset((char*)hdr + PRE_SIZE, ALLOC_FILL, size);

    return (char*)hdr + PRE_SIZE;
}

void *
js_dcalloc(size_t num, size_t size, const char *file, int line)
{
    void *ptr = js_dmalloc(size * num, file, line);
    if (ptr)
        memset(ptr, 0, size*num);

    return ptr;
}

void *
js_drealloc(void *old_ptr, size_t new_size, const char *file, int line)
{
    Header *hdr;
    void   *new_ptr;

    if (!old_ptr)
        return js_dmalloc(new_size, file, line);

    hdr = (Header*)((char*)old_ptr - PRE_SIZE);
    if (!check_block(hdr, file, line))
        return NULL;

    new_ptr = js_dmalloc(new_size, file, line);
    if (!new_ptr)
        return NULL;

    memcpy(new_ptr, old_ptr, hdr->size < new_size ? hdr->size : new_size);

    js_dfree(old_ptr, file, line);

    return new_ptr;
}

void
js_SetAllocationLimit(size_t limit)
{
    g_alloc_limit = limit;
}

int
js_CheckMemoryStatus(void)
{
    Header *hdr;
    int     result = 1;

    if (g_head != NULL) {
        result = 0;
        dlog("Memory leak: %lu bytes in %lu blocks.\n",
             g_alloc_bytes, g_alloc_blocks);
    }

    for (hdr = g_head; hdr != NULL; hdr = hdr->next) {
        dlog("Leaked block %p (%lu bytes) %s:%d\n",
             (char*)hdr + PRE_SIZE, hdr->size, hdr->alloc_file,
             hdr->alloc_line);
    }

    if (g_corrupt_blocks) {
        result = 0;
        dlog("Corrupted blocks were detected\n");
    }

    return result;
}

int
js_AllocLimitReached(void)
{
    return g_alloc_requests > g_alloc_limit;
}
